from pandas import DataFrame, RangeIndex
from eda_report.exceptions import InputError


def validate_input_dtype(data):
    """Check if the data is a pandas DataFrame, and raise an InputError if it
    isn't."""
    if isinstance(data, DataFrame):
        return data
    else:
        try:
            # Cast the data as a dataframe
            data = DataFrame(data).infer_objects()
            # Ensure the data has meaningful column names
            if isinstance(data.columns, RangeIndex):
                data.columns = [f'var_{i+1}' for i in data.columns]
            return data
        except Exception:
            raise InputError(
                f'Expected a pandas.Dataframe object, but got {type(data)}.'
            )



import pandas as pd
from eda_report.exceptions import InputError


def df_from_file(filename):
    """Load the file `filename` as a pandas DataFrame."""
    file_extension = filename.rsplit('.')[-1]

    if file_extension == 'csv':
        return pd.read_csv(filename)
    elif file_extension == 'xlsx':
        return pd.read_excel(filename, engine='openpyxl')
    else:
        raise InputError(f'Invalid input file: {filename}')



class Error(Exception):
    """Base class for exceptions in this package."""
    pass


class InputError(Error):
    """Exception raised for invalid input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message



import seaborn as sns
import numpy as np
from itertools import combinations
from eda_report.plotting import Fig, savefig
from tqdm import tqdm


class MultiVariable:
    def __init__(self, data, graph_colour='cyan'):
        self.data = data
        self.GRAPH_COLOUR = graph_colour
        self.get_bivariate_analysis()

    def get_bivariate_analysis(self):
        numeric_cols = self.data.select_dtypes(include='number').columns

        if numeric_cols.size > 1:
            self.correlation_df = self.data.corr()
            self._plot_joint_scatterplot()
            self._plot_joint_correlation()
            self._compare_variable_pairs()

    def _plot_joint_scatterplot(self):
        """Create a joint scatter-plot of all numeric columns."""
        fig = sns.pairplot(
            self.data.select_dtypes(include='number'), height=1.75,
            plot_kws={'color': self.GRAPH_COLOUR},
            diag_kws={'color': self.GRAPH_COLOUR}
        )
        fig.fig.suptitle('Scatter-plots of Numeric Columns', x=0.5, y=1.04,
                         size=20)
        self.joint_scatterplot = savefig(fig)

    def _plot_joint_correlation(self):
        """Plot a heatmap of the correlation in all numeric columns."""
        fig = Fig(figsize=(6, 6))
        ax = fig.subplots()
        sns.heatmap(self.correlation_df, annot=True, yticklabels=True,
                    mask=np.triu(self.correlation_df), ax=ax,
                    cmap=sns.light_palette(self.GRAPH_COLOUR, as_cmap=True))
        ax.tick_params(rotation=45)
        fig.suptitle('Correlation in Numeric Columns', size=15)

        self.joint_correlation_plot = savefig(fig)

    def _get_variable_pairs(self):
        """Get a list of unique pairings of the numeric variables"""
        self.var_pairs = set(combinations(self.correlation_df.columns, r=2))

    def _regression_plot(self, var1, var2):
        """Create a scatterplot with a fitted linear regression line.

        Parameters:
        ----------
        var1, var2: string
            A pair of numeric column(variable) names.
        """
        fig = Fig(figsize=(8.2, 4))
        ax1, ax2 = fig.subplots(1, 2)
        sns.regplot(x=var1, y=var2, data=self.data, ax=ax1, truncate=False,
                    color=self.GRAPH_COLOUR)
        sns.regplot(x=var2, y=var1, data=self.data, ax=ax2, truncate=False,
                    color=self.GRAPH_COLOUR)
        ax1.set_title(f'Scatterplot - {var1} vs {var2}'.title(), size=9)
        ax2.set_title(f'Scatterplot - {var2} vs {var1}'.title(), size=9)

        self.bivariate_scatterplots[(var1, var2)] = savefig(fig)

    def _quantify_correlation(self, var1, var2):
        """Explain the magnitude of correlation.

        Parameters:
        ----------
        var1, var2: string
            A pair of numeric column(variable) names.
        """
        correlation = self.correlation_df.loc[var1, var2]
        nature = ' positive' if correlation > 0 else ' negative'

        value = abs(correlation)
        if value >= 0.9:
            strength = 'very strong'
        elif value >= 0.7:
            strength = 'strong'
        elif value >= 0.5:
            strength = 'moderate'
        elif value >= 0.3:
            strength = 'weak'
        elif value >= 0.1:
            strength = 'very weak'
        else:
            strength = 'virtually no'
            nature = ''

        self.corr_type[(var1, var2)] = \
            f'{strength}{ nature} correlation ({correlation:.2f})'

    def _compare_variable_pairs(self):
        """
        Get a brief summary of the nature of correlation between pairs of
        numeric variables.
        """
        self._get_variable_pairs()
        self.corr_type = {}
        self.bivariate_scatterplots = {}

        for var1, var2 in tqdm(self.var_pairs, ncols=79):
            self._quantify_correlation(var1, var2)
            self._regression_plot(var1, var2)



from docx import Document
from docx.shared import Inches
from eda_report.univariate import Variable
from eda_report.multivariate import MultiVariable
from eda_report.validate import validate_input_dtype
from tqdm import tqdm
import logging


logging.basicConfig(
    format='[%(levelname)s %(asctime)s.%(msecs)03d] %(message)s',
    level=logging.INFO, datefmt='%H:%M:%S'
)


class ReportDocument:
    def __init__(self, data, title='Exploratory Data Analysis Report',
                 output_filename='basic-eda-report.docx',
                 graph_colour='orangered', table_style='Table Grid'):
        self.data = validate_input_dtype(data)
        self.TITLE = title
        self.GRAPH_COLOUR = graph_colour
        self.TABLE_STYLE = table_style
        self.OUTPUT_FILENAME = output_filename
        self.document = Document()
        self.get_report()

    def get_report(self):
        """Calculate statistics, plot graphs, and save results as a .docx file.
        """
        logging.info('Assessing correlation in numeric variables...')
        self.variables = MultiVariable(self.data,
                                       graph_colour=self.GRAPH_COLOUR)
        logging.info('Done. Summarising each variable...')
        self._create_title_page()  # begin the report document
        self._get_variable_info()  # summarise each variable

        if hasattr(self.variables, 'var_pairs'):
            self._get_bivariate_analysis()  # summarise variable pairs

        self._save_file()
        logging.info(f'Done. Results saved as {self.OUTPUT_FILENAME!r}')

    def _create_title_page(self):
        """Add a title and a brief summary of the data."""
        # Title
        self.document.add_heading(self.TITLE, level=0)

        # Brief summary
        num_rows, num_cols = self.data.shape
        num_numeric = self.data.select_dtypes(include='number').columns.size

        if num_rows > 1:
            rows = f'{num_rows} rows (observations)'
        else:
            rows = '1 row (observation)'

        if num_cols > 1:
            cols = f'{num_cols} columns (features)'
        else:
            cols = '1 column (feature)'

        if num_numeric > 1:
            numeric = f', {num_numeric} of which are numeric'
        elif num_numeric == 1:
            numeric = ', 1 of which is numeric'
        else:
            numeric = ''

        intro = f'The dataset consists of {rows} and {cols}{numeric}.'

        self.document.add_paragraph(intro)

        if hasattr(self.variables, 'joint_scatterplot'):
            self.document.add_picture(self.variables.joint_scatterplot,
                                      width=Inches(6.5))
            self.document.add_page_break()

    def _get_variable_info(self):
        """Get a brief introduction, summary statistics, and graphs for each
        individual variable."""
        for idx, col in enumerate(tqdm(self.data.columns, ncols=79),
                                  start=1):
            var = Variable(self.data[col], graph_colour=self.GRAPH_COLOUR)
            # Heading
            self.document.add_heading(f'{idx}. {col.title()}', level=2)
            # Introduction
            p = self.document.add_paragraph()
            p.add_run(f'{col.capitalize()}').bold = True
            p.add_run(f' is a {var.var_type} variable ')
            p.add_run(f'with {var.num_unique} unique values. ')
            p.add_run(f'{var.missing} of its values are missing.')
            # Summary statistics
            self.document.add_heading('Summary Statistics', level=4)
            self._create_table(var.statistics)
            if hasattr(var, 'most_common_items'):
                self.document.add_heading('Most Common Values', level=4)
                self._create_table(var.most_common_items)
            # Graphs
            self.document.add_picture(var.graphs, width=Inches(5))

            self.document.add_page_break()

    def _get_bivariate_analysis(self):
        """Get comparisons and scatterplots for pairs of numeric variables."""
        self.document.add_heading('Bivariate Analysis (Correlation)', level=1)
        self.document.add_paragraph()
        self.document.add_picture(self.variables.joint_correlation_plot,
                                  width=Inches(6.7))
        self.document.add_page_break()

        for idx, var_pair in enumerate(self.variables.var_pairs, start=1):
            # Heading
            self.document.add_heading(
                f'{idx}. {var_pair[0].title()} vs {var_pair[1].title()}',
                level=2)
            # Introductory text
            p = self.document.add_paragraph()
            p.add_run(f'{var_pair[0].capitalize()}').bold = True
            p.add_run(' and ')
            p.add_run(f'{var_pair[1].capitalize()}').bold = True
            p.add_run(f' have {self.variables.corr_type[var_pair]}.')
            # Scatter-plot
            self.document.add_picture(
                self.variables.bivariate_scatterplots[var_pair],
                width=Inches(6)
            )
            self.document.add_paragraph()

            if idx % 2 == 0:
                # Add page break after every 2 pairs
                self.document.add_page_break()

    def _create_table(self, data):
        """Create a table from the given data and add it to the document."""
        table = self.document.add_table(rows=len(data), cols=2)
        # Set table style
        table.style = self.document.styles[self.TABLE_STYLE]
        # Set column dimensions
        table.columns[0].width = Inches(2.5)
        table.columns[1].width = Inches(2)

        # Get a list of tuples, each to be a row in the table
        items = [(f'{label}', f'{val}') for label, val in data.items()]

        for idx, row in enumerate(table.rows):
            label, value = items[idx]
            row.cells[0].text = label
            row.cells[1].text = value
        self.document.add_paragraph()

    def _save_file(self):
        """Save the document as a .docx file."""
        self.document.save(self.OUTPUT_FILENAME)



import matplotlib
from io import BytesIO
from matplotlib.figure import Figure


# Matplotlib configuration
matplotlib.rcParams['figure.dpi'] = 150  # resolution (dots per inch)
matplotlib.rcParams['figure.autolayout'] = True  # tight layout
matplotlib.rcParams['savefig.transparent'] = False   # transparent background
matplotlib.rcParams["savefig.edgecolor"] = 'k'  # black frame
matplotlib.rcParams['font.family'] = 'serif'
matplotlib.rcParams['axes.spines.top'] = False  # remove top spine
matplotlib.rcParams['axes.spines.right'] = False  # remove left spine
matplotlib.use('agg')  # use non-interactive matplotlib back-end

Fig = Figure


def savefig(figure):
    """Save a matplotlib figure in a file-like object, and return it."""
    graph = BytesIO()
    figure.savefig(graph)
    return graph



import seaborn as sns
from eda_report.plotting import Fig, savefig
from pandas.api.types import is_numeric_dtype, is_bool_dtype


class Variable:
    """Blueprint for the structure and properties of each column/variable."""

    def __init__(self, data, graph_colour='cyan'):
        self.data = data
        self.GRAPH_COLOUR = graph_colour
        self.get_data_properties()
        self.get_summary_statictics()

    def get_data_properties(self):
        """Get basic details about the variable's data e.g type & name."""
        self.name = self.data.name
        self.num_unique = self.data.nunique()

        missing_values = self.data.isna().sum()
        if missing_values == 0:
            self.missing = "None"
        else:
            self.missing = \
                f"{missing_values} ({missing_values / len(self.data):.2%})"

        if is_numeric_dtype(self.data):
            if is_bool_dtype(self.data):
                self.var_type = 'categorical'
                self.data = self.data.astype('category')
            else:
                self.var_type = 'numeric'
        else:
            # Handle bool, string and datetime data as categorical
            self.var_type = 'categorical'

    def get_summary_statictics(self):
        """Get summary statistics for the variable as a pandas Series."""
        if self.var_type == 'numeric':
            self._numeric_summary_statictics()
            self._plot_numeric()
        elif self.var_type == 'categorical':
            self._categorical_summary_statictics()
            self._plot_categorical()

    def _numeric_summary_statictics(self):
        """Get summary statistics for a numeric variable."""
        summary = self.data.describe()
        summary.index = [
            'Number of observations', 'Average', 'Standard Deviation',
            'Minimum', 'Lower Quartile', 'Median', 'Upper Quartile', 'Maximum'
        ]
        summary['Skewness'] = self.data.skew()
        summary['Kurtosis'] = self.data.kurt()
        self.statistics = summary.round(7)

    def _categorical_summary_statictics(self):
        """Get summary statistics for a categorical variable."""
        summary = self.data.describe()[['count', 'unique', 'top']]
        summary.index = ['Number of observations', 'Unique values',
                         'Mode (Highest occurring value)']
        self.statistics = summary

        most_common_items = self.data.value_counts().head()
        n = len(self.data)
        self.most_common_items = \
            most_common_items.apply(lambda x: f'{x} ({x / n:.2%})')

    def _plot_numeric(self):
        """Get a boxplot and a histogram for a numeric variable."""
        fig = Fig(figsize=(6, 6), linewidth=1)
        ax1, ax2 = fig.subplots(2, 1)
        # Box-plot
        ax1.boxplot(self.data.dropna(), vert=False, notch=True)
        ax1.set_yticklabels([''])
        ax1.set_xlabel(f'{self.name}')
        ax1.set_title(f'Box-plot of {self.name}', size=12)
        # Histogram
        ax2.set_title(f'Distribution plot of {self.name}', size=12)
        sns.histplot(x=self.data, kde=True, ax=ax2, color=self.GRAPH_COLOUR)

        self.graphs = savefig(fig)

    def _plot_categorical(self, color='cyan'):
        """Get a bar-plot for a categorical variable."""
        fig = Fig(figsize=(6, 4), linewidth=1)
        ax = fig.subplots()

        self.data.value_counts().nlargest(10).plot.bar(
            color=self.GRAPH_COLOUR, ax=ax
        )
        ax.tick_params(axis='x', rotation=45)
        ax.set_title(f'Bar-plot of {self.name}', size=12)

        # Annotate bars
        for p in ax.patches:
            ax.annotate(f'{p.get_height():,}', ha='left',
                        xy=(p.get_x(), p.get_height()*1.02),)

        self.graphs = savefig(fig)



import argparse
from eda_report.read_file import df_from_file


def process_cli_args():
    """Process command line arguments to analyse a file."""
    parser = argparse.ArgumentParser(
        prog='eda_report',
        description='Get a basic EDA report in docx format.')

    # Get the data from a file
    parser.add_argument(
        'infile', type=df_from_file,
        help='A .csv or .xlsx file to process.')

    # Set the output file's name
    parser.add_argument(
        '-o', '--outfile', default='eda-report.docx',
        help='The output file (default: %(default)s)')

    # Set the report's title
    parser.add_argument(
        '-t', '--title', default='Exploratory Data Analysis Report',
        help='The top level heading in the report (default: %(default)s)')

    # Set the graph colour
    parser.add_argument(
        '-c', '--colour', default='orangered',
        help='A valid matplotlib color specifier (default: %(default)s)')

    # Parse the arguments
    return parser.parse_args()


if __name__ == '__main__':
    process_cli_args()



from eda_report.document import ReportDocument
from eda_report.cli import process_cli_args

__version__ = '0.0.4'


def get_word_report(data, *, title='Exploratory Data Analysis Report',
                    graph_colour='orangered',
                    output_filename='basic-report.docx'):
    """Get an EDA report in .docx format."""
    ReportDocument(data, title=title, graph_colour=graph_colour,
                   output_filename=output_filename)


def run_from_cli():
    """Create the report using arguments from the command line."""
    args = process_cli_args()
    get_word_report(args.infile, title=args.title, graph_colour=args.colour,
                    output_filename=args.outfile)
